import functools
import os
import sys
from typing import Callable
from typing import Optional

import orjson
import redis
from cachetools.func import ttl_cache as cachetools_ttl_cache
from logbook import Logger
from logbook import StreamHandler

StreamHandler(sys.stdout, level="WARNING").push_application()
logger = Logger(__name__)

_REDIS_POOL = None


class RedisCached:
    def __init__(self,
                 pool: Optional[redis.ConnectionPool],
                 ttl_seconds: int,
                 function: Callable,
                 is_method: bool = False,
                 cache_falsy: bool = True,
                 serializer: Callable = orjson.dumps,
                 deserializer: Callable = orjson.loads):

        self._redis = redis.Redis(connection_pool=pool)
        self.function = function
        self.serializer = serializer
        self.deserializer = deserializer
        self.ttl_seconds = ttl_seconds
        self.is_method = is_method
        self.cache_falsy = cache_falsy

    @property
    def key_prefix(self):
        return self.function.__name__

    def key(self, *args, **kwargs):
        if self.is_method:
            args = args[1:]
        params = self.serializer({'args': args, "kwargs": kwargs})
        return f"{self.key_prefix}__{params}"

    @property
    def __name__(self):
        return self.function.__name__

    @property
    def __wrapped__(self):
        return self.function

    def __call__(self, *args, **kwargs):
        val = None
        key = self.key(*args, **kwargs)
        try:
            val = self._redis.get(key)
        except redis.exceptions.RedisError:
            logger.exception("Unable to use cache")

        if val is not None:
            logger.debug("Cache HIT for %s", self.key(*args, **kwargs))
            return self.deserializer(val)

        logger.debug("Cache MISS for %s", self.key(*args, **kwargs))
        val = self.function(*args, **kwargs)

        if not val and not self.cache_falsy:
            logger.debug("Caching falsy result is disabled for %s", self.__name__)
            return val

        try:
            self._redis.setex(key, self.ttl_seconds, self.serializer(val))
            logger.debug("Cache SET for %s", self.key(*args, **kwargs))
        except redis.exceptions.RedisError:
            logger.exception("Unable to set cache")

        return val

    def clear_all(self):
        try:
            keys = list(self._redis.scan_iter(match=f"{self.key_prefix}*"))
            if keys:
                self._redis.delete(*keys)
        except redis.exceptions.RedisError:
            logger.exception("Unable to clear cache")
        else:
            logger.debug("Cache CLEARED for %s", keys)


def ttl_cache(ttl: int, *args, is_method: bool = True,
              cache_falsy: bool = True, **kwargs):
    global _REDIS_POOL
    redis_url = os.getenv('REDIS_URL')
    if not redis_url:
        logger.warning("REDIS_URL is not set falling back to memory cache")
        return cachetools_ttl_cache(*args, ttl=ttl, **kwargs)
    if _REDIS_POOL is None:
        logger.warning("Redis pool initializing")
        _REDIS_POOL = redis.ConnectionPool.from_url(
            redis_url, max_connections=10, socket_connect_timeout=5,
            socket_timeout=5, decode_responses=True
        )

    def decorator(func: Callable):
        cached_func = RedisCached(
            _REDIS_POOL,
            ttl,
            function=func,
            is_method=is_method,
            cache_falsy=cache_falsy
        )

        def wrapper(*wargs, **wkwargs):
            # Re-wrapping to preserve function signature
            return cached_func(*wargs, **wkwargs)

        wrapper = functools.update_wrapper(wrapper, func)
        wrapper.cache_clear = cached_func.clear_all
        return wrapper

    return decorator



"""High level API."""

from __future__ import annotations

import sys
from typing import List
from typing import Sequence
from typing import Type
from typing import Union

from logbook import Logger
from logbook import StreamHandler

from rs2wapy.adapters import PlayerWrapper
from rs2wapy.adapters import WebAdminAdapter
from rs2wapy.adapters.adapters import BanWrapper
from rs2wapy.adapters.adapters import MemberWrapper
from rs2wapy.adapters.adapters import SquadWrapper
from rs2wapy.adapters.adapters import TrackingWrapper
from rs2wapy.models import AccessPolicy
from rs2wapy.models import AllTeam
from rs2wapy.models import ChatMessage
from rs2wapy.models import CurrentGame
from rs2wapy.models import MapCycle
from rs2wapy.models import Player
from rs2wapy.models import PlayerScoreboard
from rs2wapy.models import Team
from rs2wapy.models import TeamScoreboard

StreamHandler(sys.stdout, level="WARNING").push_application()
logger = Logger(__name__)


class RS2WebAdmin:
    """Provides a high-level interface to Rising Storm 2: Vietnam
    server's WebAdmin tool.
    """

    def __init__(self, username: str, password: str, webadmin_url: str):
        """
        :param username: RS2 WebAdmin username.
        :param password: RS2 WebAdmin password.
        :param webadmin_url: RS2 WebAdmin URL.
        """
        self._adapter = WebAdminAdapter(username, password, webadmin_url)

    def get_chat_messages(self) -> Sequence[ChatMessage]:
        """Return new chat messages since last time this method
        was called and after the creation of this RS2WebAdmin instance.
        """
        return self._adapter.get_chat_messages()

    def post_chat_message(self, message: str, team: Type[Team] = AllTeam):
        """Post new chat message, visible to specific team(s).

        :param message:
            The chat message to post.
        :param team:
            The team the message is visible to.
        """
        self._adapter.post_chat_message(message, team)

    def get_current_game(self) -> CurrentGame:
        """Return object representing current game information.
        """
        return self._adapter.get_current_game()

    def change_map(self, new_map: str, url_extra: dict = None):
        """Change map.

        :param new_map:
            New map name string.
        :param url_extra:
            Dictionary, with extra URL variables as keys
            and URL variable values as values.

        Example call:
        change_map("VNTE-Resort", url_extra={
          "MaxPlayers": 64,
          "mutator": "ExampleMutator",
        })

        The url_extra parameter corresponds to the WebAdmin
        'Additional URL variables' input option.
        """
        if url_extra is None:
            url_extra = {}
        self._adapter.change_map(new_map, url_extra)

    def get_maps(self) -> dict:
        """Return maps currently installed on the server.
        Return value is a dictionary with game mode names
        as keys and map name lists as values:

        Example return value:
        {
          'ROGame.ROGameInfoTerritories': ['VNTE-Resort', 'VNTE-CuChi'],
          'ROGame.ROGameInfoSupremacy': ['VNSU-Resort'],
        }
        """
        return self._adapter.get_maps()

    def get_maps_list(self) -> List[str]:
        """Return list of all maps of all game modes
        currently installed on the server.
        """
        return self._adapter.get_maps_list()

    def get_players(self) -> List[PlayerWrapper]:
        """Return players currently online on the server.
        Return value is a list of adapters.PlayerWrapper objects
        representing the players on the server at the time of
        the invocation of this method.
        """
        return self._adapter.get_players()

    def get_player_scoreboard(self) -> PlayerScoreboard:
        """Return current player scoreboard. Player scoreboard
        does not store player IDs, because deducing them
        from WebAdmin is unreliable.
        """
        return self._adapter.get_current_game().player_scoreboard

    def get_team_scoreboard(self) -> TeamScoreboard:
        """Return current team scoreboard."""
        return self._adapter.get_current_game().team_scoreboard

    def get_squads(self) -> List[SquadWrapper]:
        """Return current squads."""
        return self._adapter.get_squads()

    def get_banned_players(self) -> List[BanWrapper]:
        """Return banned players."""
        return self._adapter.get_banned_players()

    def get_tracked_players(self) -> List[TrackingWrapper]:
        """Return tracked players.

        WARNING: This method is extremely slow for servers
        with large tracking databases.

        TODO: Can we leverage asyncio?
        TODO: Background collection like chat?
        """
        return self._adapter.get_tracked_players()

    def get_access_policies(self) -> List[AccessPolicy]:
        """Return access policies."""
        raise NotImplementedError
        # return self._adapter.get_access_policies()

    def add_access_policy(self, ip_mask: str, policy: str):
        raise NotImplementedError
        # self._adapter.add_access_policy(ip_mask, policy)

    def ban_player(self, player: Union[Player, PlayerWrapper],
                   reason: str, duration: str = None,
                   notify_players: bool = False):
        # TODO: Use correct notation when referring to
        #  "external" variables in the docstring.
        """Ban player from the server.

        :param player:
            The player to ban.
        :param reason:
            Ban reason.
        :param duration:
            Duration string. Ban is permanent if no
            duration string is supplied.

            If the string is ill-formed,
            the ban will be interpreted as permanent.

            The expected format is '{length}{ws}{unit}',
            where {length} is a positive integer, {ws}
            is an optional whitespace and {unit} is one of
            `rs2wapy.adapters.BAN_EXP_UNITS`. The string is
            case-insensitive.

            Example duration strings:
            '4 Hour'
            '3day'
            '1Year'
        :param notify_players:
            If True, notify players on the server.
        """
        self._adapter.ban_player(player, reason=reason, duration=duration,
                                 notify_players=notify_players)

    def kick_player(self, player: Union[Player, PlayerWrapper],
                    reason: str, notify_players: bool = False):
        """Kick player from the server.

        :param player:
            The player to kick.
        :param reason:
            Kick reason.
        :param notify_players:
            If True, notify players on the server.
        """
        self._adapter.kick_player(player, reason, notify_players)

    def session_ban_player(self, player: Union[Player, PlayerWrapper],
                           reason: str, notify_players: bool = False):
        """Session ban player from the server.
        Session bans reset when the server changes level.

        :param player:
            The player to session ban.
        :param reason:
            Session ban reason.
        :param notify_players:
            If True, notify players on the server.
        """
        self._adapter.session_ban_player(player, reason, notify_players)

    def get_map_cycles(self) -> List[MapCycle]:
        """Return map cycles."""
        return self._adapter.get_map_cycles()

    def set_map_cycles(self, map_cycles: List[MapCycle]):
        """Set map cycles."""
        self._adapter.set_map_cycles(map_cycles)

    def get_advertisement_messages(self) -> List[str]:
        raise NotImplementedError

    def set_advertisement_messages(self, ad_msgs: List[str]):
        raise NotImplementedError

    def get_advertisement_interval(self) -> int:
        raise NotImplementedError

    def set_advertisement_interval(self, ad_interval: int):
        raise NotImplementedError

    def get_members(self) -> List[MemberWrapper]:
        return self._adapter.get_members()



"""
Provides utilities for parsing responses from
Rising Storm 2: Vietnam WebAdmin.

TODO: This module is growing fast. Needs refactoring.
"""
from __future__ import annotations

import re
import sys
from functools import lru_cache
from typing import Dict
from typing import List
from typing import Sequence
from typing import Tuple
from typing import Union

from bs4 import BeautifulSoup
from logbook import Logger
from logbook import StreamHandler
from steam.steamid import SteamID

import rs2wapy.models as models
from rs2wapy.adapters import adapters
from rs2wapy.epicgamesstore import EGSID
from rs2wapy.steam import SteamWebAPI

StreamHandler(sys.stdout, level="WARNING").push_application()
logger = Logger(__name__)

TEAMCOLOR_PATTERN = re.compile(r"background: (.*);")
ROUND_LIMIT_SUB_PATTERN = re.compile(r"\?RoundLimit=([0-9]*)")
ROUND_LIMIT_MATCH_PATTERN = re.compile(r".*\?RoundLimit=([0-9]*).*")
NO_PLAYERS = ["There are no players"]
UNIQUE_ID_KEY = "Unique ID"
TEAM_INDEX_KEY = "\xa0"


class RS2WebAdminResponseParser:
    # TODO: Refactor magic numbers.

    def __init__(self, encoding: str = None):
        if not encoding:
            encoding = "iso-8859-1"
        self._encoding = encoding

    def parse_html(self, resp: bytes, encoding: str = None) -> BeautifulSoup:
        if not encoding:
            encoding = self._encoding
        return BeautifulSoup(resp.decode(encoding), features="html.parser")

    def parse_chat_messages(self, resp: bytes,
                            encoding: str = None) -> Sequence[models.ChatMessage]:
        parsed_html = self.parse_html(resp, encoding)
        chat_message_divs = parsed_html.find_all(
            "div", attrs={"class": "chatmessage"})
        # parsed_html.find_all("div", attrs={"class": "chatnotice"})
        cm = []
        for div in chat_message_divs:
            cm.append(self.parse_chat_message(div))
        return cm

    def parse_access_policy(self, resp: bytes,
                            encoding: str = None) -> List[str]:
        parsed_html = self.parse_html(resp, encoding)
        policy_table = parsed_html.find("table", attrs={"id": "policies"})
        trs = policy_table.find_all("tr")
        policies = []
        for tr in trs:
            ip_mask = tr.find("input", attrs={"name": "ipmask"})
            policy = tr.find("option", attrs={"selected": "selected"})
            if ip_mask and policy:
                policies.append(
                    f"{ip_mask.get('value')}: {policy.text.upper()}")
        return policies

    def parse_current_game(self, resp: bytes) -> models.CurrentGame:
        """TODO: headers vs. rows length assertion!"""
        parsed_html = self.parse_html(resp)

        info = {}
        rules = {}
        p_scores = {}
        t_scores = {}

        logger.info("parsing Ranked status")
        ranked = parsed_html.find(
            "span", attrs={"class": "ranked"}).text
        ranked = True if ranked.lower() == "ranked: yes" else False
        info["Ranked"] = ranked

        logger.info("parsing player scoreboard")
        player_scoreboard_table = parsed_html.find(
            "table", attrs={"id": "players"})
        player_scoreboard_thead = player_scoreboard_table.find("thead")
        p_thead = player_scoreboard_thead.find_all(
            "a", attrs={"class": "sortable"})
        p_headers = ["Team"]
        p_headers.extend([h.text.strip() for h in p_thead])
        p_headers.extend(["Admin", "Spectator"])

        p_row_elements = player_scoreboard_table.find_all("tr")
        p_scoreboard_parsed = self._parse_table(p_row_elements)
        for p_cols in p_scoreboard_parsed:
            p_cols[0] = "North" if p_cols[0] == "0" else "South"
            for p_key, p_col in zip(p_headers, p_cols):
                p_scores[p_key] = p_col
        player_scoreboard = models.PlayerScoreboard(stats=p_scores)

        logger.info("parsing team scoreboard")
        team_scoreboard_table = parsed_html.find(
            "table", attrs={"id": "teams"})
        t_thead = team_scoreboard_table.find("thead")
        t_headers = t_thead.find_all("th")
        t_headers = [h.text.strip() for h in t_headers]
        t_headers = ["Team Index"] + t_headers[1:]

        t_row_elements = team_scoreboard_table.find_all("tr")
        t_scoreboard_parsed = self._parse_table(t_row_elements)
        for t_cols in t_scoreboard_parsed:
            for t_key, t_col in zip(t_headers, t_cols):
                t_scores[t_key] = t_col

        team_scoreboard = models.TeamScoreboard(stats=t_scores)

        logger.info("parsing currentGame and currentRules")
        info_element = parsed_html.find("dl", attrs={"id": "currentGame"})
        rules_element = parsed_html.find("dl", attrs={"id": "currentRules"})

        info_dts = info_element.find_all("dt")
        info_dds = info_element.find_all("dd")
        rules_dts = rules_element.find_all("dt")
        rules_dds = rules_element.find_all("dd")

        len_info_dts = len(info_dts)
        len_info_dds = len(info_dds)
        len_rules_dts = len(rules_dts)
        len_rules_dds = len(rules_dds)

        logger.debug(
            "found {lidts} info_dts, {lidds} info_dds, "
            "{lrtss} rules_dts, {lrdds} rules_dds",
            lidts=len_info_dts,
            lidds=len_info_dds,
            lrtss=len_rules_dts,
            lrdds=len_rules_dds,
        )

        if len_info_dts != len_info_dds:
            logger.warning(
                "possible missing info data: len_info_dts({lidts}) "
                "!= len_info_dds({lidds})",
                lidts=len_info_dts, lidds=len_info_dds,
            )

        if len_info_dts != len_info_dds:
            logger.warning(
                "possible missing rules data: len_rules_dts({lrdts}) "
                "!= len_rules_dds({lrdds})",
                lrdts=len_rules_dts, lrdds=len_rules_dds,
            )

        logger.info("parsing Map and Game Type")
        # Get "Map" and "Game Type", which contain
        # <code> tag in their text.
        code_elements = []
        for idt, idd in zip(info_dts, info_dds):
            code_el = idd.find("code")
            if code_el:
                code_elements.append((idt, idd.find("code").text))
                idt.extract()
                idd.extract()
        for cd in code_elements:
            info[cd[0].text] = cd[1]

        # Find again after extracting some
        # elements to avoid duplicating them.
        info_dts = info_element.find_all("dt")
        info_dds = info_element.find_all("dd")

        for idt, idd in zip(info_dts, info_dds):
            info[idt.text] = idd.text
        for rdt, rdd in zip(rules_dts, rules_dds):
            rules[rdt.text] = rdd.text

        time_limit = rules["Time Limit"]
        limit, remaining, remainder = time_limit.split("seconds")
        limit = f"{limit.strip()} seconds"
        remaining = f"{remaining.strip()} seconds"
        rules["Time Limit"] = f"{limit} ({remaining} {remainder.strip()})"

        cmpgn_active = info["MP Campaign Active"]
        info["MP Campaign Active"] = (
            True if cmpgn_active.startswith("Yes") else False)

        info["Server Name"] = info["Server Name"].split("\r")[0]

        return models.CurrentGame(
            player_scoreboard=player_scoreboard,
            team_scoreboard=team_scoreboard,
            info=info,
            rules=rules,
        )

    def parse_mutator_group_count(self, resp: bytes) -> int:
        parsed_html = self.parse_html(resp)
        mutator_group_count = int(parsed_html.find(
            "input", attrs={"name": "mutatorGroupCount"}).get("value"))
        return mutator_group_count

    def parse_game_type_options(self, resp: bytes) -> List[str]:
        parsed_html = self.parse_html(resp)
        options = parsed_html.find(
            "select", attrs={"id": "gametype"}).find_all("option")
        return [o.get("value").strip() for o in options]

    def parse_map_options(self, resp: bytes) -> List[str]:
        parsed_html = self.parse_html(resp)
        options = parsed_html.find(
            "select", attrs={"id": "map"}).find_all("option")
        return [o.get("value").strip() for o in options]

    def parse_players(self, resp: bytes, adapter: adapters.WebAdminAdapter
                      ) -> List[adapters.PlayerWrapper]:
        parsed_html = self.parse_html(resp)
        if not parsed_html:
            logger.error(
                "unable to parse players; no response"
                " data to parse")
            return []

        player_table = parsed_html.find("table", attrs={"id": "players"})

        player_headers = player_table.find("thead")
        player_headers = player_headers.find_all("th")

        player_table = player_table.find("tbody")
        player_table = player_table.find_all("tr")
        player_table = self._parse_table(player_table)

        # Not expecting this header to ever change order.
        # We could probably hard-code this...
        player_headers = [ph.text for ph in player_headers]
        player_headers[player_headers.index(TEAM_INDEX_KEY)] = "Team Index"

        if (len(player_table) == 1) and player_table[0] == NO_PLAYERS:
            logger.debug("no players")
            return []

        if not all(len(p) for p in player_table):
            logger.error(
                "player rows in player table differ in length")
            return []

        if not len(player_table[0]) == len(player_headers):
            logger.error("player table and player headers differ in length")
            return []

        # We use 'Unique ID' column here instead of 'Steam ID'
        # column because 'Steam ID' column is sometimes not filled.
        # 'Unique ID' is just a hex-string of the user's SteamID64
        # anyway. EGS players are also handled in similar way.
        id_index = player_headers.index(UNIQUE_ID_KEY)

        players = []
        id_to_stats: Dict[Union[SteamID, EGSID], Dict] = {}
        steam_ids = []
        egs_ids = []

        for player_row in player_table:
            unique_id = player_row[id_index]

            try:
                int_id = int(unique_id, 16)
                is_steam_id = self._is_steam_id(int_id)
            except ValueError as ve:
                logger.exception(ve)
                logger.error("invalid Unique ID: '{uid}', skipping", uid=unique_id)
                continue

            stats = {
                key: value for key, value in zip(
                    player_headers, player_row)
                if key.lower() != "actions"
            }

            if is_steam_id:
                steam_id = SteamID(int_id)
                steam_ids.append(steam_id)
                id_to_stats[steam_id] = stats
            else:
                egs_id = EGSID(int_id)
                egs_ids.append(egs_id)
                id_to_stats[egs_id] = stats

        persona_names = SteamWebAPI().get_persona_names(
            steam_ids=steam_ids
        )

        for steam_id in steam_ids:
            persona_name = ""
            try:
                persona_name = persona_names[steam_id]
            except KeyError as ke:
                logger.error(
                    "error getting persona name for Steam ID: {sid}",
                    sid=steam_id)
                logger.exception(ke)

            p_stats = id_to_stats[steam_id]
            player = models.Player(ident=steam_id, stats=p_stats, persona_name=persona_name)

            players.append(adapters.PlayerWrapper(
                player=player,
                adapter=adapter,
            ))

        for egs_id in egs_ids:
            p_stats = id_to_stats[egs_id]

            player = models.Player(ident=egs_id, stats=p_stats)

            players.append(adapters.PlayerWrapper(
                player=player,
                adapter=adapter,
            ))

        return players

    def parse_hash_alg(self, resp: bytes) -> str:
        parsed_html = self.parse_html(resp)
        login_script = parsed_html.find("script", text=re.compile(r".*hashAlg.*"))
        match = re.search(r'.*var\shashAlg\s=\s"(.*)";.*', str(login_script))
        alg = ""
        if match:
            alg = match.group(1)
        return alg

    def parse_map_list_indices(self, resp) -> dict:
        parsed_html = self.parse_html(resp)
        map_list_idxs = parsed_html.find(
            "select", attrs={"id": "maplistidx"})
        map_list_idxs = map_list_idxs.find_all("option")

        valid_idxs = {}
        for mli in map_list_idxs:
            try:
                idx = int(mli.get("value"))
                if idx >= 0:
                    is_active = "active" in mli.text.lower()
                    valid_idxs[idx] = is_active
            except ValueError as ve:
                logger.warning("{ve} during map list index conversion to int",
                               ve=ve)

        return valid_idxs

    def parse_map_cycle(self, resp) -> List[Tuple[str, int]]:
        """
        Return list of tuples (map name, round limit).
        """
        parsed_html = self.parse_html(resp)
        maps = parsed_html.find(
            "textarea", attrs={"id": "mapcycle"}
        ).text
        maps = maps.split("\n")

        round_limits = []
        for m in maps:
            rl = re.match(ROUND_LIMIT_MATCH_PATTERN, m)
            if rl:
                round_limits.append(int(rl.group(1)))
            else:
                round_limits.append(0)

        if len(round_limits) != len(maps):
            logger.error("round limit list and map list length mismatch")

        maps = [(re.sub(ROUND_LIMIT_SUB_PATTERN, "", m), rl)
                for m, rl in zip(maps, round_limits)]
        return maps

    def parse_squads(self, resp: bytes,
                     adapter: adapters.WebAdminAdapter
                     ) -> List[adapters.SquadWrapper]:
        parsed_html = self.parse_html(resp)
        parsed_html = parsed_html.find("table", attrs={"id": "squads"})

        squads_table = parsed_html.find("tbody")
        squads_table = squads_table.find_all("tr")
        squads_table = self._parse_table(squads_table)

        squads = []
        for row in squads_table:
            try:
                team_idx = int(row[0].strip())
                team = models.Team.from_team_index(team_idx)
            except Exception as e:
                team = models.UnknownTeam
                logger.error("unable to parse squad team: {e}", e=e)

            number = -1
            try:
                number = int(row[1])
            except Exception as e:
                logger.error("unable to parse squad number: {e}", e=e)

            name = ""
            try:
                name = row[2]
            except Exception as e:
                logger.error("unable to parse squad name: {e}", e=e)

            squads.append(models.Squad(
                team=team,
                number=number,
                name=name,
            ))

        squad_wrappers = [
            adapters.SquadWrapper(
                squad=squad,
                adapter=adapter,
            ) for squad in squads
        ]

        return squad_wrappers

    def parse_tracking(self, resp: bytes,
                       adapter: adapters.WebAdminAdapter
                       ) -> List[adapters.TrackingWrapper]:

        parsed_html = self.parse_html(resp)
        tracking_table = parsed_html.find("table", attrs={"id": "tracking"})
        tracking_headers = [th.text for th in tracking_table.find_all("th")]
        tracking_tbody = tracking_table.find("tbody")
        tb_row_entries = tracking_tbody.find_all("tr")
        tracking_rows = self._parse_table(tb_row_entries)

        # Unique ID field is more reliable as Steam ID
        # might be missing in the table in some cases.
        id_index = tracking_headers.index(UNIQUE_ID_KEY)

        id_to_tracking_data = {}
        steam_ids = []
        egs_ids = []

        for row in tracking_rows:
            unique_id = row[id_index]

            try:
                int_id = int(unique_id, 16)
                is_steam_id = self._is_steam_id(int_id)
            except ValueError as ve:
                logger.exception(ve)
                logger.error("invalid Unique ID: '{uid}', skipping", uid=unique_id)
                continue

            tracking_data = {
                key: value for key, value in zip(
                    tracking_headers, row)
                if key.lower() != "actions"
            }

            if is_steam_id:
                steam_id = SteamID(int_id)
                steam_ids.append(steam_id)
                id_to_tracking_data[steam_id] = tracking_data
            else:
                egs_id = EGSID(int_id)
                egs_ids.append(egs_id)
                id_to_tracking_data[egs_id] = tracking_data

        persona_names = SteamWebAPI().get_persona_names(
            steam_ids=steam_ids
        )

        tracking_wrappers = []

        for steam_id in steam_ids:
            persona_name = ""
            try:
                persona_name = persona_names[steam_id]
            except KeyError as ke:
                logger.error(
                    "error getting persona name for Steam ID: {sid}",
                    sid=steam_id)
                logger.exception(ke)

            tracking_data = id_to_tracking_data[steam_id]
            player = models.Player(ident=steam_id, persona_name=persona_name)

            tracking_wrappers.append(adapters.TrackingWrapper(
                player=player,
                tracking_data=tracking_data,
                adapter=adapter,
            ))

        for egs_id in egs_ids:
            tracking_data = id_to_tracking_data[egs_id]

            player = models.Player(ident=egs_id)

            tracking_wrappers.append(adapters.TrackingWrapper(
                player=player,
                tracking_data=tracking_data,
                adapter=adapter,
            ))

        return tracking_wrappers

    def parse_bans(self, resp: bytes,
                   adapter: adapters.WebAdminAdapter
                   ) -> List[adapters.BanWrapper]:

        parsed_html = self.parse_html(resp)
        ban_table = parsed_html.find("table", attrs={"class": "grid"})
        ban_headers = [th.text for th in ban_table.find_all("th")]
        ban_tbody = ban_table.find("tbody")
        ban_row_entries = ban_tbody.find_all("tr")
        ban_rows = self._parse_table(ban_row_entries)

        # Unique ID field is more reliable as Steam ID
        # might be missing in the table in some cases.
        id_index = ban_headers.index(UNIQUE_ID_KEY)

        id_to_extra_data = {}
        steam_ids = []
        egs_ids = []

        for row in ban_rows:
            unique_id = row[id_index]

            try:
                int_id = int(unique_id, 16)
                is_steam_id = self._is_steam_id(int_id)
            except ValueError as ve:
                logger.exception(ve)
                logger.error("invalid Unique ID: '{uid}', skipping", uid=unique_id)
                continue

            extra_data = {
                key: value for key, value in zip(
                    ban_headers, row)
                if key.lower() != "actions"
            }

            if is_steam_id:
                steam_id = SteamID(int_id)
                steam_ids.append(steam_id)
                id_to_extra_data[steam_id] = extra_data
            else:
                egs_id = EGSID(int_id)
                egs_ids.append(egs_id)
                id_to_extra_data[egs_id] = extra_data

        persona_names = SteamWebAPI().get_persona_names(
            steam_ids=steam_ids
        )

        ban_wrappers = []

        for steam_id in steam_ids:
            persona_name = ""
            try:
                persona_name = persona_names[steam_id]
            except KeyError as ke:
                logger.error(
                    "error getting persona name for Steam ID: {sid}",
                    sid=steam_id)
                logger.exception(ke)

            extra_data = id_to_extra_data[steam_id]
            player = models.Player(ident=steam_id, persona_name=persona_name)

            ban = models.Ban(
                player=player,
                reason=extra_data["Reason"],
                until=extra_data["Expires On"],
                admin=extra_data["Admin"],
                when=extra_data["When"],
            )

            ban_wrappers.append(adapters.BanWrapper(
                ban=ban,
                adapter=adapter,
            ))

        for egs_id in egs_ids:
            extra_data = id_to_extra_data[egs_id]

            player = models.Player(ident=egs_id)

            ban = models.Ban(
                player=player,
                reason=extra_data["Reason"],
                until=extra_data["Expires On"],
                admin=extra_data["Admin"],
                when=extra_data["When"],
            )

            ban_wrappers.append(adapters.BanWrapper(
                ban=ban,
                adapter=adapter,
            ))

        return ban_wrappers

    def parse_fvri(self, resp: bytes):
        """Parse first visible row index from response."""
        parsed_html = self.parse_html(resp)
        return parsed_html.find(
            "input", attrs={"id": "__FirstVisibleRowIndex"}).get("value")

    def parse_has_next_page(self, resp: bytes):
        """Return True if next page button is enabled."""
        parsed_html = self.parse_html(resp)
        np_button = parsed_html.find(
            "button", attrs={"id": "__NextPage"})
        if not np_button:
            return False
        button_disabled = np_button.has_attr("disabled")
        return not button_disabled

    def parse_player_id_key(
            self, resp: bytes,
            player: Union[models.Player, adapters.PlayerWrapper]
    ) -> Tuple[int, str]:
        """Parse "hidden" player key and ID from players table."""
        parsed_html = self.parse_html(resp)
        trs = parsed_html.find_all("tr")
        player_id = -1
        player_key = ""

        for tr in trs:
            div = tr.find("div")
            if not div:
                continue

            # Find x from <input id = "__PlayerId_x">.
            x = div.find("input").get("id").split("_")[-1]
            div_player_key = div.find(
                "input", attrs={"id": f"__PlayerKey_{x}"}).get("value")
            div_player_name = div.find(
                "input", attrs={"id": f"__PlayerName_{x}"}).get("value")

            div_player_steam_id = div_player_key.split("_")[1].strip().lower()
            steam_id = player.steam_id.as_64
            # Match leading zeros of strings to compare.
            steam_id = f"{steam_id:#0{len(div_player_steam_id)}x}"
            steam_id = steam_id.strip().lower()

            if (steam_id == div_player_steam_id
                    and (div_player_name == player.name)):
                player_id = int(div.find(
                    "input", attrs={"id": f"__PlayerId_{x}"}).get("value"))
                player_key = div_player_key
                break

        return player_id, player_key

    def parse_members(self, resp: bytes,
                      adapter: adapters.WebAdminAdapter
                      ) -> List[adapters.MemberWrapper]:
        parsed_html = self.parse_html(resp)
        members_table = parsed_html.find("table", attrs={"id": "members"})
        members_headers = [th.text for th in members_table.find_all("th")]
        tracking_tbody = members_table.find("tbody")
        tb_row_entries = tracking_tbody.find_all("tr")
        members_rows = self._parse_table(tb_row_entries)

        id_index = members_headers.index(UNIQUE_ID_KEY)

        id_to_extra_data = {}
        steam_ids = []
        egs_ids = []

        for row in members_rows:
            unique_id = row[id_index]

            try:
                int_id = int(unique_id, 16)
                is_steam_id = self._is_steam_id(int_id)
            except ValueError as ve:
                logger.exception(ve)
                logger.error("invalid Unique ID: '{uid}', skipping", uid=unique_id)
                continue

            tracking_data = {
                key: value for key, value in zip(
                    members_headers, row)
                if key.lower() != "actions"
            }

            if is_steam_id:
                steam_id = SteamID(int_id)
                steam_ids.append(steam_id)
                id_to_extra_data[steam_id] = tracking_data
            else:
                egs_id = EGSID(int_id)
                egs_ids.append(egs_id)
                id_to_extra_data[egs_id] = tracking_data

        persona_names = SteamWebAPI().get_persona_names(
            steam_ids=steam_ids
        )

        member_wrappers = []

        for steam_id in steam_ids:
            persona_name = ""
            try:
                persona_name = persona_names[steam_id]
            except KeyError as ke:
                logger.error(
                    "error getting persona name for Steam ID: {sid}",
                    sid=steam_id)
                logger.exception(ke)

            extra_data = id_to_extra_data[steam_id]
            player = models.Player(ident=steam_id, persona_name=persona_name)

            member_wrappers.append(adapters.MemberWrapper(
                player=player,
                member_data=extra_data,
                adapter=adapter,
            ))

        for egs_id in egs_ids:
            extra_data = id_to_extra_data[egs_id]

            player = models.Player(ident=egs_id)

            member_wrappers.append(adapters.MemberWrapper(
                player=player,
                member_data=extra_data,
                adapter=adapter,
            ))

        return member_wrappers

    @staticmethod
    def parse_chat_message(div: BeautifulSoup) -> models.ChatMessage:
        teamcolor = str(div.find(
            "span", attrs={"class": "teamcolor"}).get("style"))
        if not teamcolor:
            logger.error(
                "no teamcolor in chat message div={div}", div=div)
        else:
            try:
                match = re.match(TEAMCOLOR_PATTERN, teamcolor)
                if match:
                    teamcolor = match.groups()[0]
            except IndexError as ie:
                logger.error("error getting teamcolor: {e}", e=ie)

        teamnotice = div.find("span", attrs={"class": "teamnotice"})
        if teamnotice:
            teamnotice = teamnotice.text

        name = div.find("span", attrs={"class": "username"})
        if name:
            name = name.text

        msg = div.find("span", attrs={"class": "message"})
        if msg:
            msg = msg.text

        return models.ChatMessage(
            sender=name,
            text=msg,
            team=models.Team.from_hex_color(str(teamcolor)),
            channel=models.ChatChannel.from_teamnotice(teamnotice)
        )

    @staticmethod
    def _parse_table(row_elements: Sequence) -> List[List[str]]:
        all_cols = []
        for row in row_elements:
            cols = row.find_all("td")
            cols = [ele.get_text(" ").strip() for ele in cols]
            if not cols:
                continue
            all_cols.append(cols)
        return all_cols

    @staticmethod
    @lru_cache(maxsize=64)
    def _is_steam_id(unique_id: Union[str, int]) -> bool:
        """Best-effort check whether hex-string or integer is Steam ID.

        :param unique_id: Unique ID hex-string or integer.
        :return: True if Steam ID.
        """
        if isinstance(unique_id, str):
            try:
                unique_id = int(unique_id, 16)
            except ValueError:
                raise ValueError(f"unable to convert Unique ID "
                                 f"'{unique_id}' to int")

        # Very likely a Steam ID.
        return len(str(unique_id)) == 17

    @staticmethod
    @lru_cache(maxsize=64)
    def _id_to_player(unique_id: str) -> models.Player:
        raise NotImplementedError



from typing import Union


class EGSID:

    def __init__(self, ident: Union[int, str]):
        if isinstance(ident, str):
            ident = int(ident, 16)
        self._ident = ident

    @property
    def ident(self) -> int:
        return self._ident

    @ident.setter
    def ident(self, ident: int):
        self._ident = ident



from __future__ import annotations

import os
import sys
from typing import Dict
from typing import List
from typing import Sequence

import requests
import steam.webapi
from cachetools import TTLCache
from cachetools import cached
from logbook import Logger
from logbook import StreamHandler
from steam import steamid

StreamHandler(sys.stdout, level="WARNING", bubble=True).push_application()
logger = Logger(__name__)

_ttl_cache = TTLCache(maxsize=256, ttl=60)


def _chunks(seq: Sequence, n: int):
    """Yield successive n-sized chunks from seq."""
    for i in range(0, len(seq), n):
        yield seq[i:i + n]


class Singleton(type):
    _instances: Dict[type, Singleton] = {}

    def __call__(cls, *args, **kwargs):
        try:
            steam_api_key = os.environ["STEAM_WEB_API_KEY"]
            if cls not in cls._instances:
                cls._instances[cls] = super().__call__(
                    steam_api_key, *args, **kwargs)
        except KeyError as ke:
            logger.info("'STEAM_WEB_API_KEY' environment variable not set, "
                        "some features are not available")
            logger.debug(ke)
            cls._instances[cls] = super().__call__(*args, dummy=True, **kwargs)
        except requests.exceptions.HTTPError as e:
            logger.debug(e, exc_info=True)
            logger.warning("unable to initialize Steam Web API, "
                           "some features are not available")
            cls._instances[cls] = super().__call__(*args, dummy=True, **kwargs)
        return cls._instances[cls]


class SteamWebAPI(steam.webapi.WebAPI, metaclass=Singleton):
    """Helper class for using Steam Web API quickly."""

    _REQUESTS_MADE = 0

    @property
    def requests_made(self) -> int:
        """Return the number of requests made to Steam API."""
        return self._REQUESTS_MADE

    def __init__(self, *args, dummy=False, **kwargs):
        self._dummy = dummy
        if not dummy:
            super().__init__(*args, **kwargs)

    @cached(cache=_ttl_cache)
    def get_persona_name(self, steam_id: steamid.SteamID) -> str:
        # TODO: Refer to variable in docstring.
        """Return persona name for Steam ID.
        Use get_persona_names for multiple requests to limit
        the number of requests made to Steam API.
        """
        if self._dummy:
            return ""
        else:
            # noinspection PyUnresolvedReferences
            response = self.ISteamUser.GetPlayerSummaries(
                steamids=steam_id.as_64)["response"]
            ret = response["players"][0]["personaname"]
            SteamWebAPI._REQUESTS_MADE += 1
            return ret

    def get_persona_names(self, steam_ids: List[steamid.SteamID]
                          ) -> Dict[steamid.SteamID, str]:
        """Return dictionary of Steam IDs to persona names
        for given Steam IDs. Queries the Steam Web API in
        batches of 100 Steam IDs.
        """
        if self._dummy:
            return {steam_id: "" for steam_id in steam_ids}

        ret = {}

        for steam_id in steam_ids:
            try:
                personaname = _ttl_cache[steam_id]
                ret[steam_id] = personaname
            except KeyError:
                pass

        for chunk in _chunks(steam_ids, n=100):
            chunk_ids = [str(cid.as_64)
                         for cid in chunk
                         if cid not in ret]
            chunk_ids_str = ",".join(chunk_ids)

            # noinspection PyUnresolvedReferences
            resp = self.ISteamUser.GetPlayerSummaries(
                steamids=chunk_ids_str)["response"]["players"]
            SteamWebAPI._REQUESTS_MADE += 1

            for r in resp:
                steam_id = steamid.SteamID(r["steamid"])
                personaname = r["personaname"]
                ret[steam_id] = personaname
                _ttl_cache[steam_id] = personaname

            input_len = len(chunk_ids)
            output_len = len(ret)
            num_bad = int(abs(input_len - output_len))
            if input_len != output_len:
                logger.warn(
                    f"Steam API did not return valid value "
                    f"for {num_bad} input Steam IDs "
                    f"(input_len={input_len}, output_len={output_len})")

        return ret



from .adapters import BanWrapper
from .adapters import PlayerWrapper
from .adapters import SquadWrapper
from .adapters import WebAdminAdapter
from .adapters import logger

__all__ = [
    "BanWrapper",
    "WebAdminAdapter",
    "SquadWrapper",
    "PlayerWrapper",
    "logger",
]
